from typing import Tuple, List

import torch
from torch import nn

from .deep_memory import MemoryState, FastLinearMemory
from .base import get_rng_state
class Feedforward(nn.Module):
    """
    A classic feedforward implementation.

    d_model is the input and output dimension.
    d_hidden is how wide we get inside.
    """

    def __init__(self,
                 d_model: int,
                 d_hidden: int
                 ):
        super().__init__()
        self.d_hidden = d_hidden
        self.d_model = d_model

        self.ff1 = nn.Linear(d_model, d_hidden)
        self.ff2 = nn.Linear(d_hidden, d_model)
        self.relu = nn.ReLU()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        tensor = self.ff1(tensor)
        tensor = self.relu(tensor)
        tensor = self.ff2(tensor)
        return tensor


class DecoderLayer(nn.Module):
    """
    A deep transformer decoder layer

    Consists of the deep memory unit,
    and the feedforward block. Can
    operate in both forward and reverse
    modes.
    """

    def __init__(self,
                 d_model: int,
                 d_hidden: int,
                 d_address: int,
                 d_memory: int,
                 num_read_heads: int,
                 num_write_heads: int,
                 num_memories: int,
                 numeric_write_factor: float,
                 dropout: float,
                 device: torch.device,
                 dtype: torch.dtype
                 ):
        """
        :param d_model: The main model dimensions, as the embeddings come in as
        :param d_hidden: The dimensions of the feedforward process
        :param d_address: The dimensions used for memory addressing
        :param d_memory: The dimensions used for memory storage
        :param num_read_heads: The number of memory read heads
        :param num_write_heads: The numbery of memory write heads
        :param num_memories: The number of discrete memories
        :param dropout: The dropout probability
        :param device: The device we are on
        :param dtype: The dtype we are on
        """
        super().__init__()

        self.d_model = d_model
        self.d_hidden = d_hidden
        self.d_address = d_address
        self.d_memory = d_memory
        self.num_read_heads = num_read_heads
        self.num_write_heads = num_write_heads
        self.num_memories = num_memories
        self.dropout_rate = dropout

        # Define the deep memory layer

        self.deep_memories = FastLinearMemory(d_model,
                                              d_address,
                                              d_memory,
                                              num_read_heads,
                                              num_write_heads,
                                              num_memories,
                                              dropout,
                                              max_write_factor=numeric_write_factor,
                                              device=device,
                                              dtype=dtype)
        self.deep_layernorm = nn.LayerNorm(d_model)

        # Define the feedforward layer
        self.feedforward = Feedforward(d_model, d_hidden)
        self.ff_layernorm = nn.LayerNorm(d_model)

        # Define the dropout layer
        self.dropout = nn.Dropout(dropout)

    def create_state(self, batch_shape: torch.Size) -> MemoryState:
        """
        Creates a blank memory state associated with the
        given batch size.
        :param batch_shape: The batch shape to match
        :return: The setup memory state
        """
        return self.deep_memories.create_state(batch_shape)

    def reverse(self,
                tensor: torch.Tensor,
                batch_mask: torch.Tensor,
                next_memory: MemoryState
                ) -> Tuple[Tuple[torch.Tensor, MemoryState], MemoryState]:
        """
        The reverse mechanism. Able to perform the same computation,
        but with the notable complication of also returning the
        prior memory state, setup to accumulate gradients.

        :param tensor: The tensor input to use
        :param batch_mask: Shape (...). Indicates whether memory can be updated. True means No.
        :param next_memory: The next memory state during the forward pass
        :return:
        -Tuple:
            - The original output, but with a graph on it.
            - The previous memory state .
        """

        # Perform the deep memory access pattern


        (update, next_memory), previous_memory = self.deep_memories.reverse(tensor, batch_mask, next_memory)
        tensor = self.deep_layernorm(tensor + self.dropout(update))


        # Perform the feedforward
        update = self.feedforward(tensor)
        tensor = self.ff_layernorm(tensor + self.dropout(update))

        # Return
        return (tensor, next_memory), previous_memory

    def forward(self,
                tensor: torch.Tensor,
                batch_mask: torch.Tensor,
                previous_memory: MemoryState) -> Tuple[torch.Tensor, MemoryState]:
        """
        The forward mechanism. Performs a forward pass through the mode
        :param tensor: The input.
        :param batch_mask: Shape (...). Indicates whether memory can be updated. True means No.
        :param previous_memory: The current memory state.
        :return:
            - The output.
            - The next memory state.
        """
        # Perform the deep memory access pattern


        update, next_memory = self.deep_memories(tensor, batch_mask, previous_memory)
        tensor = self.deep_layernorm(tensor + self.dropout(update))


        # Perform the feedforward
        update = self.feedforward(tensor)
        tensor = self.ff_layernorm(tensor + self.dropout(update))

        # Return
        return tensor, next_memory


class RecurrentDecoder(nn.Module):
    """
    A recurrent decoder transformer mechanism, with
    forward and reverse modes.

    In forward mode it should be provided with an input embedding
    and the memory states from the previous layer. In
    reverse mode, it should be provided with an input embedding,
    and the memory states from the next layer. Either way, it
    computes a functioning graph.

    However, during the backwards pass it returns the prior
    memory states, rather than the next ones, and sets them
    up to accumulate gradients.

    Bottlenecks are built at the recurrent decoder level.
    This makes it quite easy to adjust and pretrain the model
    for different embeddings or different embedding widths.
    You can fetch the old decoder layers feature, and use
    it to initialize a new recurrent decoder.
    """

    def __init__(self,
                 d_model: int,
                 dropout_rate: float,
                 decoder_layers: List[DecoderLayer]
                 ):

        super().__init__()
        template = decoder_layers[0]
        self.d_model = d_model
        self.dropout_rate = dropout_rate

        # Setup layernorms, with a no op on the last for immediate consumption
        # by a logit system.
        layernorms = [nn.LayerNorm(d_model) for _ in range(len(decoder_layers))]
        layernorms.pop(-1)
        layernorms.append(nn.Identity())

        # Setup bottleneck, unbottleneck layers
        bottlenecks = [nn.Linear(d_model, template.d_model) for _ in range(len(decoder_layers))]
        unbottlenecks = [nn.Linear(template.d_model, d_model) for _ in range(len(decoder_layers))]

        # Setup layers and repositories
        self.layernorms = nn.ModuleList(layernorms)
        self.decoder_layers = nn.ModuleList(decoder_layers)
        self.dropout = nn.Dropout(dropout_rate)

        self.bottlenecks = nn.ModuleList(bottlenecks)
        self.unbottlenecks = nn.ModuleList(unbottlenecks)

    def rebuild_at_different_width(self, d_model: int) -> 'RecurrentDecoder':
        """
        Rebuilds an existing model to use a different d_model. Expect
        to do a moderate amount of fine tuning, but most of the core
        logic should still be useful.

        :param d_model: The model width to now set us up at
        :return: The new recurrent decoder
        """
        return RecurrentDecoder(d_model, self.dropout_rate, self.decoder_layers)

    def create_state(self, batch_shape: torch.Size) -> List[MemoryState]:
        """
        Sets up the recurrent state bound
        to a particular batch shape

        :param batch_shape: The batch shape to match
        :return: A list of memory states. One for each layer.
        """
        states = []
        for decoder_layer in self.decoder_layers:
            states.append(decoder_layer.create_state(batch_shape))
        return states

    def reverse(self,
                embedding: torch.Tensor,
                batch_mask: torch.Tensor,
                next_memories: List[MemoryState]
                ) -> Tuple[Tuple[torch.Tensor, List[MemoryState]], List[MemoryState]]:
        """
        Runs the reverse process. This means figuring out the
        previous memory states and setting them up for gradient
        accumulation. And of course returning the final output

        :param embedding: The input embedding. Whatever it might be
        :param batch_mask: Shape (...). Indicates whether memory can be updated. True means No.
        :param next_memories: The memories from the NEXT step
        :return:
        - Tuple:
            - The final embedding, ready for usage in logits.
            - The memory states for this timestep. It has a graph. We need to insert gradients here
        - The memory from the last timestep. Setup to accumulate gradients and continue the chain.
        """
        assert embedding.shape[-1] == self.d_model
        previous_memories = []
        gradprop_memories = []
        iterator = zip(self.layernorms, self.decoder_layers, self.bottlenecks, self.unbottlenecks, next_memories)
        for layernorm, decoder_layer, bottleneck, unbottleneck, next_memory in iterator:
            # Bottleneck, process, and unbottleneck
            bottlenecked_embedding = bottleneck(embedding)
            (bottlenecked_embedding, next_memory), previous_memory = decoder_layer.reverse(bottlenecked_embedding,
                                                                                              batch_mask,
                                                                                              next_memory)

            update = unbottleneck(bottlenecked_embedding)

            # Stash away memories
            gradprop_memories.append(next_memory)
            previous_memories.append(previous_memory)

            # Integrate update, and store memory.
            embedding = layernorm(embedding + self.dropout(update))
        return (embedding, gradprop_memories), previous_memories

    def forward(self,
                embedding: torch.Tensor,
                batch_mask: torch.Tensor,
                previous_memories: List[MemoryState]
                ) -> Tuple[torch.Tensor, List[MemoryState]]:
        """
        The forward mechanism. Performs a forward pass through the model.
        This will usually occur without gradients.
        :param embedding: The embedding being processed
        :param batch_mask: Shape (...). Indicates whether memory can be updated. True means No.
        :param previous_memories: The memory states from the last timestep
        :return:
        - The output. Same whether forward or backwards
        - The memory states for the next timestep.
        """
        assert embedding.shape[-1] == self.d_model
        next_memories = []
        iterator = zip(self.layernorms, self.decoder_layers, self.bottlenecks, self.unbottlenecks, previous_memories)
        for layernorm, decoder_layer, bottleneck, unbottleneck, previous_memory in iterator:
            # Bottleneck, process, then unbottleneck
            bottlenecked_embedding = bottleneck(embedding)
            bottlenecked_embedding, next_memory = decoder_layer(bottlenecked_embedding,
                                                                batch_mask,
                                                                previous_memory)
            update = unbottleneck(bottlenecked_embedding)

            # Integrate update, then append memory
            embedding = layernorm(embedding + self.dropout(update))
            next_memories.append(next_memory)
        return embedding, next_memories


def build_decoder(
        # Primary specifications
        d_model: int,
        num_layers: int,

        # Helper specifics
        d_core: int,
        d_hidden: int,
        d_address: int,
        d_memory: int,
        num_read_heads: int,
        num_write_heads: int,
        num_memories: int,
        dropout_rate: float,
        auxilary_dropout_rate: float,
        numeric_write_factor: float,

        # Dtype, device
        dtype: torch.dtype = None,
        device: torch.device = None
) -> RecurrentDecoder:
    """
    Creates a functioning recurrent decoder, with
    forward and reverse modes, ready for integration
    into a broader architecture.

    The returned model is designed to be entirely
    recurrent.

    --- Top level parameters ---
    :param d_model:
    - Dimensionality of the embeddings coming in, and being returned
    - Most of the computation does not occur at this width.
    :param num_layers:
    - Number of transformer layers.
    - Each is a DecoderLayer
    :param dropout_rate:
    - Dropout rate, within the primary recurrent model
    :param auxilary_dropout_rate:
    - Dropout rate, within the core decoder layers and computation models.

    --- DecoderLayer Parameters
    Everything here is occurring with respect to d_core.

    :param d_core:
    - Bottleneck rate. d_model is bottlenecked down to this dimension to save compute time
    - Much of the computation occurs at this rate
    :param d_hidden:
    - Size of the hidden layer.
    - Choose it with respect to d_core.
    :param d_address:
    - A smaller subset of d_core. Can be anything.
    - Represents the width of the attn memory addresses.
    :param d_memory:
    - Represents the width of the memory value addresses.
    - Can be different.
    :param num_read_heads:
    - Number of heads used for the memory read process
    - Larger means more of the memory can be read from per step
    :param num_write_heads:
    - Number of heads used for the memory write process
    - Larger means more of the memory can be written to per step
    :param num_memories:
    - Number of independent memory states

    ---- final ---
    :param dtype: The dtype
    :param device: The device
    :return: A setup RecurrentDecoder
    """

    # Create the decoder layer stack
    layers = [
        DecoderLayer(d_core,
                     d_hidden,
                     d_address,
                     d_memory,
                     num_read_heads,
                     num_write_heads,
                     num_memories,
                     numeric_write_factor,
                     auxilary_dropout_rate,
                     device=device,
                     dtype=dtype
                     )
        for _ in range(num_layers)
    ]

    # Create and return the model
    return RecurrentDecoder(d_model, dropout_rate, layers)



import os
import textwrap
from typing import Dict, Union

import torch
from torch import nn
from transformers import PreTrainedModel, AutoTokenizer, PreTrainedTokenizer, AutoModelForCausalLM
from enum import Enum


class SpecialTokens(Enum):
    bos_token = "<BOS>"
    eos_token = "<EOS>"
    pad_token = "<PAD>"

class AdditionalSpecialTokens(Enum):
    prompt_token = "<PROMPT>"
    beginning_of_response_token = "<RESPONSE>"
    start_grid_data = "<STARTGRID>"
    end_grid_data = "<ENDGRID>"
    grid_line = "<GRIDLINE>"

class VocabularyStruct(nn.Module):
    """
    A centralized place in which vocabulary,
    embeddings, and logits can be kept and
    otherwise managed.
    """
    @classmethod
    def auto_load_from_pretrained(cls, name: str) -> 'VocabularyStruct':
        """
        Loads the vocabulary and corrosponding tokenizer
        directly from huggingface.
        :param name: The name to load from
        :return: The setup vocabulary struct
        """
        tokenizer = AutoTokenizer.from_pretrained(name)
        model = AutoModelForCausalLM.from_pretrained(name)
        return cls.load_from_pretrained(tokenizer, model)

    @classmethod
    def load_from_pretrained(cls,
                             tokenizer: PreTrainedTokenizer,
                             model: PreTrainedModel
                             ) -> 'VocabularyStruct':
        """
        Loads a vocabulary struct from a pretrained language model.
        This allows us to use a premade embedding and manipulation
        system, saving some time.

        Be aware that this will fetch the logit endpoint too. Make
        sure it is correctly configured!
        :param tokenizer: The tokenizer associated with the model
        :param model: The model to fetch off of. Should have
            - Input embedding system
            - Output logit system
        :return: The setup vocabulary struct
        """
        embeddings = model.get_input_embeddings()
        logits = model.get_output_embeddings()
        assert logits is not None, "attempted to load model without predictive capacities"
        assert embeddings is not None, "attempted to load model without embeddings"
        return cls(embeddings, tokenizer, logits)

    def customize_vocabulary(self):
        """
        Customizes the provided collection of pretrained tokenizer,
        embeddings, and logits in order to support the declared
        required special tokens
        """
        # Generate the update dictionary, and integrate it
        # into the tokenizer
        original_size = self.tokenizer.vocab_size
        special_tokens = {case.name : case.value for case in SpecialTokens}
        special_tokens["additional_special_tokens"] = [case.value for case in AdditionalSpecialTokens]
        num_added_tokens = self.tokenizer.add_special_tokens(special_tokens, replace_additional_special_tokens=False)

        # Modify the tokenizer defaults slightly
        self.tokenizer.padding_side = "right"
        self.tokenizer.true_vocab_size = original_size + num_added_tokens

        # Expand the embeddings. We take the current embeddings, extend it
        # a bit, and initialize the extension.
        with torch.no_grad():
            old_num_embeddings, embedding_dim = self.embeddings.weight.size()
            new_embeddings = nn.Embedding(old_num_embeddings + num_added_tokens, embedding_dim)
            new_embeddings.weight[:old_num_embeddings] = self.embeddings.weight

        self.embeddings = new_embeddings

        # Expand the logits to be able to predict the extra dimensions.
        with torch.no_grad():
            num_logits, d_embedding = self.logit_projector.weight.size()
            if self.logit_projector.bias is None:
                new_logits = nn.Linear(d_embedding, num_logits + num_added_tokens, bias=False)
                new_logits.weight[:num_logits] = self.logit_projector.weight
            else:
                new_logits = nn.Linear(d_embedding, num_logits)
                new_logits.weight[:num_logits] = self.logit_projector.weight
                new_logits.bias[:num_logits] = self.logit_projector.bias
        self.logit_projector = new_logits
    def save_pretrained_vocabulary(self, directory: Union[str, os.PathLike]):
        """
        Saves the vocabulary in its current configuration
        :param directory: The directory to save in.
        """
        self.tokenizer.save_pretrained(directory)
        torch.save(self.embeddings, os.path.join(directory, "embeddings.pt"))
        torch.save(self.logit_projector, os.path.join(directory, "logit_projector.pt"))

    @classmethod
    def load_pretrained_vocabulary(cls, directory: Union[str, os.PathLike])->'VocabularyStruct':
        """
        Loads a pretrained vocabulary from the indicated save directory
        :param directory: The directory to save in
        :return: The pretrained vocabulary struct
        """
        tokenizer = AutoTokenizer.from_pretrained(directory)
        embeddings = torch.load(os.path.join(directory, "embeddings.pt"))
        logit_projector = torch.load(os.path.join(directory, "logit_projector.pt"))
        return cls(embeddings, tokenizer, logit_projector)

    def __reduce__(self):
        msg = """
        Attempt was made to invoke torch.save or pickle directly. This 
        is not supported. Instead, use built in save methods on the classes.
        """
        msg = textwrap.dedent(msg)
        raise NotImplementedError(msg)

    def __init__(self,
                 embeddings: nn.Embedding,
                 tokenizer: PreTrainedTokenizer,
                 logit: nn.Linear,
                 ):
        super().__init__()

        assert embeddings.weight.shape[0] == logit.weight.shape[0], "Embeddings and logits did not have same model dim"

        self.d_model = embeddings.weight.shape[-1]
        self.embeddings = embeddings
        self.tokenizer = tokenizer
        self.logit_projector = logit
        self.customize_vocabulary()



import torch
import functools
from torch import nn
from .base import parallel_pytree_map, middle_quantiles_mean
from typing import Any, List


class BatchCollectiveReductiveGradNorm:
    """
    Performs collective gradient normalization on a group of gradients flowing through
    the class, on a per-batch basis.

    Steps:
    1. Reduce each gradient collection per batch.
    2. Find the signal that is largest per batch.
    3. Rescale all gradients in a batch to be smaller if they exceed the threshold.

    This effectively reduces all gradients flowing from one recurrent cell to another
    by the same amount, per batch.
    """

    def __init__(self,
                 num_batch_dims: int = 1,
                 rescale_threshold: float = 1.0,
                 reduction_mode: str = 'quartiles_mean',
                 verbose: bool = False,
                 ):
        """
        Initializes the gradient normalization class.

        :param num_batch_dims: The number of batch dimensions.
        :param rescale_threshold: Threshold for triggering rescaling.
        :param reduction_mode: Reduction method for gradient tensors;
              options:
              - "mean": Reduces using mean as option
              - "max": Reduces using max as an option
              - "sum": Reduces using sum as an option
              - "quartiles_mean": Performs a mean around only the middle two quantiles. Default
        """
        assert num_batch_dims >= 0
        assert reduction_mode in ['mean', 'max', 'sum', 'quartiles_mean']
        assert rescale_threshold >= 0.0

        self.num_batch_dims = num_batch_dims
        self.rescale_threshold = rescale_threshold
        self.reduction_mode = reduction_mode
        self.verbose = verbose

    def collect_gradients(self,
                          grads: torch.Tensor,
                          accumulator: List[torch.Tensor]
                          ):
        """
        Collects gradients and appends them to an accumulator.
        :param grads: Gradients to be processed.
        :param accumulator: List to store reduced gradients for later processing.
        """
        if grads.dim() < self.num_batch_dims:
            raise ValueError("Received fewer gradient dimensions than batch dimensions.")
        grads = grads.abs()
        grads = grads.flatten(self.num_batch_dims, -1)
        accumulator.append(grads)

    @staticmethod
    def rescale_gradients(grads: torch.Tensor, rescale_factor: torch.Tensor) -> torch.Tensor:
        """
        Rescales gradients by the given rescale factor.

        :param grads: Gradients to rescale.
        :param rescale_factor: Rescale factor per batch.
        :return: Updated gradients.
        """
        while rescale_factor.dim() < grads.dim():
            rescale_factor = rescale_factor.unsqueeze(-1)
        return rescale_factor * grads

    def __call__(self, grad_tree: Any) -> Any:
        """
        Processes the gradient collection for scaling as necessary.

        :param grad_tree: The gradient tree to handle.
        :return: The normalized gradient tree.
        """
        # Accumulate reduced gradients from all parts of the gradient tree
        grad_accumulator = []
        reduce_gradients = functools.partial(self.collect_gradients, accumulator=grad_accumulator)
        parallel_pytree_map(reduce_gradients, grad_tree)

        # Calculate reduced statistics per batch over entire collection.
        summaries = []
        for observation in grad_accumulator:
            if self.reduction_mode == "mean":
                summary = observation.mean(dim=-1)
            elif self.reduction_mode == "max":
                summary = observation.max(dim=-1).values
            elif self.reduction_mode == "sum":
                summary = observation.sum(dim=-1)
            elif self.reduction_mode == "quartiles_mean":
                summary = middle_quantiles_mean(observation, dim=-1)
            else:
                raise ValueError(f"Unrecognized reduction mode: {self.reduction_mode}")
            summaries.append(summary)
        observed_grads, _ = torch.stack(summaries, dim=-1).max(dim=-1)

        if self.verbose:
            print(observed_grads)

        # Compute rescale factors per batch
        requires_rescale = observed_grads > self.rescale_threshold
        rescale_factor = torch.where(requires_rescale, self.rescale_threshold / observed_grads, 1.0)

        # Rescale gradients across the tree based on batch-specific factors
        rescale_gradients = functools.partial(self.rescale_gradients, rescale_factor=rescale_factor)
        return parallel_pytree_map(rescale_gradients, grad_tree)


class BatchCollectiveQuantileClipping:
    """
    A extremum clipping algorithm that is resistant
    to outliers. It takes the mean within the
    quantiles around the mean, then clips anything
    that exceeds that mean by a factor.

    It operates on a provided pytree
    """

    def __init__(self,
                 num_batch_dims: int = 1,
                 clip_factor: float = 1000.0,
                 protection_threshold: float = 0.0001,
                 mean_mode: str = 'quantiles_mean',
                 verbose: bool = False,
                 ):
        """
        Initializes the clipping algorithm
        :param num_batch_dims: The number of batch dimensions. Each batch is clipped separately
        :param clip_factor: When the gradient exceeds the mean by this factor, clip.
        :param protection_threshold: Gradients that are smaller than this cannot be clipped.
        :param mean_mode: Either 'mean' or 'quantiles_mean'
        - mean: The mean is taken. It is then used to determine when to clip
        - quantiles_mean: The middle two quantiles are used to compute the mean instead. Default.
        """
        assert clip_factor >= 1.0
        assert mean_mode in ['mean', 'quantiles_mean']
        self.num_batch_dims = num_batch_dims
        self.mean_mode = mean_mode
        self.clip_factor = clip_factor
        self.protection_threshold = protection_threshold
        self.verbose = verbose

    def collect_gradients(self,
                          grads: torch.Tensor,
                          accumulator: List[torch.Tensor]
                          ):
        """
        Collects gradients and appends them to an accumulator.
        :param grads: Gradients to be processed.
        :param accumulator: List to store reduced gradients for later processing.
        """
        if grads.dim() < self.num_batch_dims:
            raise ValueError("Received fewer gradient dimensions than batch dimensions.")
        grads = grads.abs()
        grads = grads.flatten(self.num_batch_dims, -1)
        accumulator.append(grads)

    @staticmethod
    def clip_gradients(grads: torch.Tensor, clip_threshold: torch.Tensor) -> torch.Tensor:
        """
        Performs the actual gradient clipping process.
        :param grads: The gradients to clip.
        :param clip_threshold: The threshold to clip to if needed.
        :return: The clipped gradients
        """
        while clip_threshold.dim() < grads.dim():
            clip_threshold = clip_threshold.unsqueeze(-1)
        return torch.minimum(grads, clip_threshold)

    def __call__(self, grad_tree: Any) -> Any:
        """
        Runs the clipping algorithm.
        :param grad_tree: The gradient tree to clip
        :return: The clipped gradients
        """
        # Collect the gradients, so we can find the clip values
        accumulator = []
        collect_gradients = functools.partial(self.collect_gradients, accumulator=accumulator)
        parallel_pytree_map(collect_gradients, grad_tree)

        # Calculate clip threshold.
        observations = torch.concat(accumulator, dim=-1)

        if self.mean_mode == "mean":
            observed_grads = observations.mean(dim=-1)
        elif self.mean_mode == "quantiles_mean":
            observed_grads = middle_quantiles_mean(observations, dim=-1)
        else:
            raise ValueError(f"Unrecognized reduction mode: {self.mean_mode}")

        threshold = observed_grads * self.clip_factor
        threshold = threshold.clip(min=self.protection_threshold)
        if self.verbose:
            print(threshold)

        # Perform any required clipping.
        clip_gradients = functools.partial(self.clip_gradients, clip_threshold=threshold)
        return parallel_pytree_map(clip_gradients, grad_tree)


class GradClip:
    """
    A simple gradient clipper.

    Gradients that go above a certain threshold
    are clipped away.
    """

    def __init__(self, clip_threshold: float):
        self.clip_threshold = clip_threshold

    def clip_gradients(self, grad: torch.Tensor) -> torch.Tensor:
        return grad.clip(max=self.clip_threshold, min=-self.clip_threshold)

    def __call__(self, grad_tree: Any) -> Any:
        """
        Runs the grad clipping process
        :param grad_tree: A grad tree to clip
        :return: The clipped gradients
        """
        return parallel_pytree_map(self.clip_gradients, grad_tree)
